import os
from itertools import product, chain
from pathlib import Path

import numpy as np
from typing import Literal, Iterable
from inspect import getfullargspec
from warnings import warn

import pygfx

from rendercanvas import BaseRenderCanvas

from ._utils import (
    make_canvas_and_renderer,
    create_controller,
    create_camera,
    get_extents_from_grid,
)
from ._utils import controller_types as valid_controller_types
from ._subplot import Subplot
from ._engine import GridLayout, WindowLayout, ScreenSpaceCamera
from .. import ImageGraphic
from ..tools import Tooltip


class Figure:
    def __init__(
        self,
        shape: tuple[int, int] = (1, 1),
        rects: list[tuple | np.ndarray] = None,
        extents: list[tuple | np.ndarray] = None,
        cameras: (
            Literal["2d", "3d"]
            | Iterable[Iterable[Literal["2d", "3d"]]]
            | pygfx.PerspectiveCamera
            | Iterable[Iterable[pygfx.PerspectiveCamera]]
        ) = "2d",
        controller_types: (
            Iterable[Iterable[Literal["panzoom", "fly", "trackball", "orbit"]]]
            | Iterable[Literal["panzoom", "fly", "trackball", "orbit"]]
        ) = None,
        controller_ids: (
            Literal["sync"]
            | Iterable[int]
            | Iterable[Iterable[int]]
            | Iterable[Iterable[str]]
        ) = None,
        controllers: pygfx.Controller | Iterable[Iterable[pygfx.Controller]] = None,
        canvas: str | BaseRenderCanvas | pygfx.Texture = None,
        renderer: pygfx.WgpuRenderer = None,
        canvas_kwargs: dict = None,
        size: tuple[int, int] = (500, 300),
        names: list | np.ndarray = None,
        show_tooltips: bool = False,
    ):
        """
        Create a Figure containing Subplots.

        Parameters
        ----------
        shape: tuple[int, int], default (1, 1)
            shape [n_rows, n_cols] that defines a grid of subplots

        rects: list of tuples or arrays
            list of rects (x, y, width, height) that define the subplots.
            rects can be defined in absolute pixels or as a fraction of the canvas.
            If width & height <= 1 the rect is assumed to be fractional.
            Conversely, if width & height > 1 the rect is assumed to be in absolute pixels.
            width & height must be > 0. Negative values are not allowed.

        extents: list of tuples or arrays
            list of extents (xmin, xmax, ymin, ymax) that define the subplots.
            extents can be defined in absolute pixels or as a fraction of the canvas.
            If xmax & ymax <= 1 the extent is assumed to be fractional.
            Conversely, if xmax & ymax > 1 the extent is assumed to be in absolute pixels.
            Negative values are not allowed. xmax - xmin & ymax - ymin must be > 0.

            If both ``rects`` and  ``extents`` are provided, then ``rects`` takes precedence over ``extents``, i.e.
            ``extents`` is ignored when ``rects`` are also provided.

        cameras: "2d", "3d", list of "2d" | "3d", Iterable of camera instances, or Iterable of "2d" | "3d", optional
            | if str, one of ``"2d"`` or ``"3d"`` indicating 2D or 3D cameras for all subplots
            | Iterable/list/array of ``2d`` and/or ``3d`` that specifies the camera type for each subplot
            | Iterable/list/array of pygfx.PerspectiveCamera instances

        controller_types: str, Iterable, optional
            list/array that specifies the controller type for each subplot.
            Valid controller types: "panzoom", "fly", "trackball", "orbit".
            If not specified a default controller is chosen based on the camera type.
            Orthographic projections, i.e. "2d" cameras, use a "panzoom" controller by default.
            Perspective projections with a FOV > 0, i.e. "3d" cameras, use a "fly" controller by default.

        controller_ids: str, list of int, np.ndarray of int, or list with sublists of subplot str names, optional
            | If `None` a unique controller is created for each subplot
            | If "sync" all the subplots use the same controller

            This allows custom assignment of controllers

            | Example with integers:
            | sync first 2 plots, and sync last 2 plots: [[0, 0, 1], [2, 3, 3]]
            | Example with str subplot names:
            | list of lists of subplot names, each sublist is synced: [[subplot_a, subplot_b, subplot_e], [subplot_c, subplot_d]]
            | this syncs subplot_a, subplot_b and subplot_e together; syncs subplot_c and subplot_d together

        controllers: pygfx.Controller | list[pygfx.Controller] | np.ndarray[pygfx.Controller], optional
            Directly provide pygfx.Controller instances(s). Useful if you want to use a ``Controller`` from an existing
            subplot or a ``Controller`` you have already instantiated. Also useful if you want to provide a custom
            ``Controller`` subclass. Other controller kwargs, i.e. ``controller_types`` and ``controller_ids``
            are ignored if `controllers` are provided.

        canvas: str, BaseRenderCanvas, pygfx.Texture
            Canvas to draw the figure onto, usually auto-selected based on running environment.

        renderer: pygfx.Renderer, optional
            pygfx renderer instance

        canvas_kwargs: dict, optional
            kwargs to pass to the canvas

        size: (int, int), optional
            starting size of canvas in absolute pixels, default (500, 300)

        names: list or array of str, optional
            subplot names

        show_tooltips: bool, default False
            show tooltips on graphics

        """

        if rects is not None:
            if not all(isinstance(v, (np.ndarray, tuple, list)) for v in rects):
                raise TypeError(
                    f"rects must a list of arrays, tuples, or lists of rects (x, y, w, h), you have passed: {rects}"
                )
            n_subplots = len(rects)
            layout_mode = "rect"
            extents = [None] * n_subplots

        elif extents is not None:
            if not all(isinstance(v, (np.ndarray, tuple, list)) for v in extents):
                raise TypeError(
                    f"extents must a list of arrays, tuples, or lists of extents (xmin, xmax, ymin, ymax), "
                    f"you have passed: {extents}"
                )
            n_subplots = len(extents)
            layout_mode = "extent"
            rects = [None] * n_subplots

        else:
            if not all(isinstance(v, (int, np.integer)) for v in shape):
                raise TypeError(
                    f"shape argument must be a tuple[n_rows, n_cols], you have passed: {shape}"
                )
            n_subplots = shape[0] * shape[1]
            layout_mode = "grid"

            # create fractional extents from the grid
            extents = get_extents_from_grid(shape)
            # empty rects
            rects = [None] * n_subplots

        if names is not None:
            # user has specified subplot names
            subplot_names = np.asarray(names).flatten()
            # make an array without nones for sanity checks
            subplot_names_without_nones = subplot_names[subplot_names != np.array(None)]

            # make sure all names are unique
            if (
                subplot_names_without_nones.size
                != np.unique(subplot_names_without_nones).size
            ):
                raise ValueError(
                    f"subplot `names` must be unique, you have provided: {names}"
                )

            # check that there are enough subplots given the number of names
            if subplot_names.size > n_subplots:
                raise ValueError(
                    f"must provide same number or fewer subplot `names` than number of supblots specified by shape, "
                    f"extents, or rects."
                    f"You have specified {n_subplots} subplots, but {subplot_names.size} subplot names."
                )

            if subplot_names.size < n_subplots:
                # pad the subplot names with nones
                subplot_names = np.concatenate(
                    [
                        subplot_names,
                        np.asarray([None] * (n_subplots - subplot_names.size)),
                    ]
                )
        else:
            # no user specified subplot names
            if layout_mode == "grid":
                # make names that show the [row index, col index]
                subplot_names = np.asarray(
                    list(map(str, product(range(shape[0]), range(shape[1]))))
                )
            else:
                subplot_names = None

        if canvas_kwargs is not None:
            if size not in canvas_kwargs.keys():
                canvas_kwargs["size"] = size
        else:
            canvas_kwargs = {"size": size, "max_fps": 60.0, "vsync": True}

        canvas, renderer = make_canvas_and_renderer(
            canvas, renderer, canvas_kwargs=canvas_kwargs
        )

        canvas.add_event_handler(self._fpl_reset_layout, "resize")

        if isinstance(cameras, str):
            # create the array representing the views for each subplot in the grid
            cameras = np.array([cameras] * n_subplots)

        # list/tuple -> array if necessary
        cameras = np.asarray(cameras).flatten()

        if cameras.size != n_subplots:
            raise ValueError(
                f"Number of cameras: {cameras.size} does not match the number of specified subplots: {n_subplots}"
            )

        # create the cameras
        subplot_cameras = np.empty(n_subplots, dtype=object)
        for index in range(n_subplots):
            subplot_cameras[index] = create_camera(camera_type=cameras[index])

        # if controller instances have been specified for each subplot
        if controllers is not None:
            # one controller for all subplots
            if isinstance(controllers, pygfx.Controller):
                controllers = [controllers] * n_subplots

            # individual controller instance specified for each subplot
            else:
                # I found that this is better than list(*chain(<list/array>)) because chain doesn't give the right
                # result we want for arrays
                for item in controllers:
                    if isinstance(item, pygfx.Controller):
                        pass
                    elif all(isinstance(c, pygfx.Controller) for c in item):
                        pass
                    else:
                        raise TypeError(
                            f"controllers argument must be a single pygfx.Controller instance, or a Iterable of "
                            f"pygfx.Controller instances. You have passed: {controllers}"
                        )

            subplot_controllers: np.ndarray[pygfx.Controller] = np.asarray(
                controllers
            ).flatten()
            if not subplot_controllers.size == n_subplots:
                raise ValueError(
                    f"number of controllers passed must be the same as the number of subplots specified "
                    f"by shape, extents, or rects: {n_subplots}. You have passed: {subplot_controllers.size} controllers"
                ) from None

            for index in range(n_subplots):
                subplot_controllers[index].add_camera(subplot_cameras[index])

        # parse controller_ids and controller_types to make desired controller for each subplot
        else:
            if controller_ids is None:
                # individual controller for each subplot
                controller_ids = np.arange(n_subplots)

            elif isinstance(controller_ids, str):
                if controller_ids == "sync":
                    # this will end up creating one controller to control the camera of every subplot
                    controller_ids = np.zeros(n_subplots, dtype=int)
                else:
                    raise ValueError(
                        f"`controller_ids` must be one of 'sync', an array/list of subplot names, or an array/list of "
                        f"integer ids. You have passed: {controller_ids}.\n"
                        f"See the docstring for more details."
                    )

            # list controller_ids
            elif isinstance(controller_ids, (list, np.ndarray)):
                ids_flat = list(chain(*controller_ids))

                # list of str of subplot names, convert this to integer ids
                if all([isinstance(item, str) for item in ids_flat]):
                    if subplot_names is None:
                        raise ValueError(
                            "must specify subplot `names` to use list of str for `controller_ids`"
                        )

                    # make sure each controller_id str is a subplot name
                    if not all([n in subplot_names for n in ids_flat]):
                        raise KeyError(
                            f"all `controller_ids` strings must be one of the subplot names. You have passed "
                            f"the following `controller_ids`:\n{controller_ids}\n\n"
                            f"and the following subplot names:\n{subplot_names}"
                        )

                    if len(ids_flat) > len(set(ids_flat)):
                        raise ValueError(
                            f"id strings must not appear twice in `controller_ids`: \n{controller_ids}"
                        )

                    # initialize controller_ids array
                    ids_init = np.arange(n_subplots)

                    # set id based on subplot position for each synced sublist
                    for row_ix, sublist in enumerate(controller_ids):
                        for name in sublist:
                            ids_init[subplot_names == name] = -(
                                row_ix + 1
                            )  # use negative numbers to avoid collision with positive numbers from np.arange

                    controller_ids = ids_init

                # integer ids
                elif all([isinstance(item, (int, np.integer)) for item in ids_flat]):
                    controller_ids = np.asarray(controller_ids).flatten()
                    if controller_ids.max() < 0:
                        raise ValueError(
                            f"if passing an integer array of `controller_ids`, "
                            f"all the integers must be positive:{controller_ids}"
                        )

                else:
                    raise TypeError(
                        f"list argument to `controller_ids` must be a list of `str` or `int`, "
                        f"you have passed: {controller_ids}"
                    )

            if controller_ids.size != n_subplots:
                raise ValueError(
                    f"Number of controller_ids: {controller_ids.size} "
                    f"does not match the number of subplots: {n_subplots}"
                )

            if controller_types is None:
                # `create_controller()` will auto-determine controller for each subplot based on defaults
                controller_types = np.array(["default"] * n_subplots)

            # valid controller types
            if isinstance(controller_types, str):
                controller_types = np.array([controller_types] * n_subplots)

            controller_types: np.ndarray[pygfx.Controller] = np.asarray(
                controller_types
            ).flatten()
            # str controller_type or pygfx instances
            valid_str = list(valid_controller_types.keys()) + ["default"]

            # make sure each controller type is valid
            for controller_type in controller_types:
                if controller_type is None:
                    continue

                if controller_type not in valid_str:
                    raise ValueError(
                        f"You have passed the invalid `controller_type`: {controller_type}. "
                        f"Valid `controller_types` arguments are:\n {valid_str}"
                    )

            # make the real controllers for each subplot
            subplot_controllers = np.empty(shape=n_subplots, dtype=object)
            for cid in np.unique(controller_ids):
                cont_type = controller_types[controller_ids == cid]
                if np.unique(cont_type).size > 1:
                    raise ValueError(
                        "Multiple controller types have been assigned to the same controller id. "
                        "All controllers with the same id must use the same type of controller."
                    )

                cont_type = cont_type[0]

                # get all the cameras that use this controller
                cams = subplot_cameras[controller_ids == cid].ravel()

                if cont_type == "default":
                    # hacky fix for now because of how `create_controller()` works
                    cont_type = None
                _controller = create_controller(
                    controller_type=cont_type, camera=cams[0]
                )

                subplot_controllers[controller_ids == cid] = _controller

                # add the other cameras that go with this controller
                if cams.size > 1:
                    for cam in cams[1:]:
                        _controller.add_camera(cam)

        self._canvas = canvas
        self._renderer = renderer

        if layout_mode == "grid":
            n_rows, n_cols = shape
            grid_index_iterator = list(product(range(n_rows), range(n_cols)))
            self._subplots: np.ndarray[Subplot] = np.empty(shape=shape, dtype=object)
            resizeable = False

        else:
            self._subplots: np.ndarray[Subplot] = np.empty(
                shape=n_subplots, dtype=object
            )
            resizeable = True

        for i in range(n_subplots):
            camera = subplot_cameras[i]
            controller = subplot_controllers[i]

            if subplot_names is not None:
                name = subplot_names[i]
            else:
                name = None

            subplot = Subplot(
                parent=self,
                camera=camera,
                controller=controller,
                canvas=canvas,
                renderer=renderer,
                name=name,
                rect=rects[i],
                extent=extents[i],  # figure created extents for grid layout
                resizeable=resizeable,
            )

            if layout_mode == "grid":
                row_ix, col_ix = grid_index_iterator[i]
                self._subplots[row_ix, col_ix] = subplot
            else:
                self._subplots[i] = subplot

        if layout_mode == "grid":
            self._layout = GridLayout(
                self.renderer,
                subplots=self._subplots,
                canvas_rect=self.get_pygfx_render_area(),
                shape=shape,
            )

        elif layout_mode == "rect" or layout_mode == "extent":
            self._layout = WindowLayout(
                self.renderer,
                subplots=self._subplots,
                canvas_rect=self.get_pygfx_render_area(),
            )

        # underlay render pass
        self._underlay_camera = ScreenSpaceCamera()
        self._underlay_scene = pygfx.Scene()

        for subplot in self._subplots.ravel():
            self._underlay_scene.add(subplot.frame._world_object)

        # overlay render pass
        self._overlay_camera = ScreenSpaceCamera()
        self._overlay_scene = pygfx.Scene()

        # tooltip in overlay render pass
        self._tooltip_manager = Tooltip()
        self._overlay_scene.add(self._tooltip_manager.world_object)

        self._show_tooltips = show_tooltips

        self._animate_funcs_pre: list[callable] = list()
        self._animate_funcs_post: list[callable] = list()

        self._current_iter = None

        self._sidecar = None

        self._output = None

        self._pause_render = False

    @property
    def shape(self) -> list[tuple[int, int, int, int]] | tuple[int, int]:
        """Only for grid layouts of subplots: [n_rows, n_cols]"""
        if isinstance(self.layout, GridLayout):
            return self.layout.shape

    @property
    def layout(self) -> WindowLayout | GridLayout:
        """
        Layout engine
        """
        return self._layout

    @property
    def canvas(self) -> BaseRenderCanvas:
        """The canvas this Figure is drawn onto"""
        return self._canvas

    @property
    def renderer(self) -> pygfx.WgpuRenderer:
        """The renderer that renders this Figure"""
        return self._renderer

    @property
    def controllers(self) -> np.ndarray[pygfx.Controller]:
        """controllers, read-only array, access individual subplots to change a controller"""
        controllers = np.asarray([subplot.controller for subplot in self], dtype=object)

        if isinstance(self.layout, GridLayout):
            controllers = controllers.reshape(self.shape)

        controllers.flags.writeable = False
        return controllers

    @property
    def cameras(self) -> np.ndarray[pygfx.Camera]:
        """cameras, read-only array, access individual subplots to change a camera"""
        cameras = np.asarray([subplot.camera for subplot in self], dtype=object)

        if isinstance(self.layout, GridLayout):
            cameras = cameras.reshape(self.shape)

        cameras.flags.writeable = False
        return cameras

    @property
    def names(self) -> np.ndarray[str]:
        """subplot names, read-only array, access individual subplots to change a name"""
        names = np.asarray([subplot.name for subplot in self])

        if isinstance(self.layout, GridLayout):
            names = names.reshape(self.shape)

        names.flags.writeable = False
        return names

    @property
    def tooltip_manager(self) -> Tooltip:
        """manage tooltips"""
        return self._tooltip_manager

    @property
    def show_tooltips(self) -> bool:
        """show/hide tooltips for all graphics"""
        return self._show_tooltips

    @show_tooltips.setter
    def show_tooltips(self, val: bool):
        self._show_tooltips = val

        if val:
            # register all graphics
            for subplot in self:
                for graphic in subplot.graphics:
                    self._tooltip_manager.register(graphic)

        elif not val:
            self._tooltip_manager.unregister_all()

    def _render(self, draw=True):
        # draw the underlay planes
        self.renderer.render(self._underlay_scene, self._underlay_camera, flush=False)

        # With new pygfx' blending, the depth buffer is only cleared after each flush, we need a manual depth
        # clear to erase the depth values set by the underlay.
        if hasattr(self.renderer, "clear"):
            self.renderer.clear(depth=True)

        # call the animation functions before render
        self._call_animate_functions(self._animate_funcs_pre)
        for subplot in self:
            subplot._render()

        # overlay render pass
        if hasattr(self.renderer, "clear"):
            self.renderer.clear(depth=True)
        self.renderer.render(self._overlay_scene, self._overlay_camera, flush=False)

        self.renderer.flush()

        # call post-render animate functions
        self._call_animate_functions(self._animate_funcs_post)

        if draw:
            # needs to be here else events don't get processed
            self.canvas.request_draw()

    def _start_render(self):
        """start render cycle"""
        self.canvas.request_draw(self._render)

    def show(
        self,
        autoscale: bool = True,
        maintain_aspect: bool = None,
        axes_visible: bool = True,
        sidecar: bool = False,
        sidecar_kwargs: dict = None,
    ):
        """
        Begins the rendering event loop and shows the Figure, returns the canvas

        Parameters
        ----------
        autoscale: bool, default ``True``
            autoscale the Scene

        maintain_aspect: bool, default ``True``
            maintain aspect ratio

        axes_visible: bool, default ``True``
            show axes

        sidecar: bool, default ``True``
            display plot in a ``jupyterlab-sidecar``, only in jupyter

        sidecar_kwargs: dict, default ``None``
            kwargs for sidecar instance to display plot
            i.e. title, layout

        Returns
        -------
        BaseRenderCanvas
            In Qt or GLFW, the canvas window containing the Figure will be shown.
            In jupyter, it will display the plot in the output cell or sidecar.
        """

        # show was already called, return canvas
        if self._output:
            return self._output

        self._start_render()

        if sidecar_kwargs is None:
            sidecar_kwargs = dict()

        # flip y-axis if ImageGraphics are present
        for subplot in self:
            for g in subplot.graphics:
                if isinstance(g, ImageGraphic):
                    subplot.camera.local.scale_y *= -1
                    break

        if autoscale:
            for subplot in self:
                if maintain_aspect is None:
                    _maintain_aspect = subplot.camera.maintain_aspect
                else:
                    _maintain_aspect = maintain_aspect
                subplot.auto_scale(maintain_aspect=maintain_aspect)

        # set axes visibility if False
        if not axes_visible:
            for subplot in self:
                subplot.axes.visible = False

        # parse based on canvas type
        if self.canvas.__class__.__name__ == "JupyterRenderCanvas":
            if sidecar:
                from sidecar import Sidecar
                from IPython.display import display

                self._sidecar = Sidecar(**sidecar_kwargs)
                self._output = self.canvas
                with self._sidecar:
                    return display(self.canvas)
            self._output = self.canvas
            return self._output

        elif self.canvas.__class__.__name__ == "QRenderCanvas":
            self._output = self.canvas
            self._output.show()
            return self.canvas

        elif self.canvas.__class__.__name__ == "OffscreenRenderCanvas":
            # for test and docs gallery screenshots
            self._fpl_reset_layout()
            for subplot in self:
                subplot.axes.update_using_camera()

                # render call is blocking only on github actions for some reason,
                # but not for rtd build, this is a workaround
                # for CI tests, the render call works if it's in test_examples
                # but it is necessary for the gallery images too so that's why this check is here
                if "RTD_BUILD" in os.environ.keys():
                    if os.environ["RTD_BUILD"] == "1":
                        self._render()

        else:  # assume GLFW
            self._output = self.canvas

        # return the canvas
        return self._output

    def close(self):
        self._output.close()
        if self._sidecar:
            self._sidecar.close()

    def _call_animate_functions(self, funcs: list[callable]):
        for fn in funcs:
            try:
                if len(getfullargspec(fn).args) > 0:
                    fn(self)
                else:
                    fn()
            except (ValueError, TypeError):
                warn(
                    f"Could not resolve argspec of {self.__class__.__name__} animation function: {fn}, "
                    f"calling it without arguments."
                )
                fn()

    def add_animations(
        self,
        *funcs: callable,
        pre_render: bool = True,
        post_render: bool = False,
    ):
        """
        Add function(s) that are called on every render cycle.
        These are called at the Figure level.

        Parameters
        ----------
        *funcs: callable(s)
            function(s) that are called on each render cycle

        pre_render: bool, default ``True``, optional keyword-only argument
            if true, these function(s) are called before a render cycle

        post_render: bool, default ``False``, optional keyword-only argument
            if true, these function(s) are called after a render cycle

        """
        for f in funcs:
            if not callable(f):
                raise TypeError(
                    f"all positional arguments to add_animations() must be callable types, you have passed a: {type(f)}"
                )
            if pre_render:
                self._animate_funcs_pre += funcs
            if post_render:
                self._animate_funcs_post += funcs

    def remove_animation(self, func):
        """
        Removes the passed animation function from both pre and post render.

        Parameters
        ----------
        func: callable
            The function to remove, raises a error if it's not registered as a pre or post animation function.

        """
        if func not in self._animate_funcs_pre and func not in self._animate_funcs_post:
            raise KeyError(
                f"The passed function: {func} is not registered as an animation function. These are the animation "
                f" functions that are currently registered:\n"
                f"pre: {self._animate_funcs_pre}\n\npost: {self._animate_funcs_post}"
            )

        if func in self._animate_funcs_pre:
            self._animate_funcs_pre.remove(func)

        if func in self._animate_funcs_post:
            self._animate_funcs_post.remove(func)

    def clear(self):
        """Clear all Subplots"""
        for subplot in self:
            subplot.clear()

    def export_numpy(self, rgb: bool = False) -> np.ndarray:
        """
        Export a snapshot of the Figure as numpy array.

        Parameters
        ----------
        rgb: bool, default ``False``
            if True, use alpha blending to return an RGB image.
            if False, returns an RGBA array

        Returns
        -------
        np.ndarray
            [n_rows, n_cols, 3] for RGB or [n_rows, n_cols, 4] for RGBA
        """
        snapshot = self.renderer.snapshot()

        if rgb:
            bg = np.zeros(snapshot.shape).astype(np.uint8)
            bg[:, :, -1] = 255

            img_alpha = snapshot[..., -1] / 255

            rgb = snapshot[..., :-1] * img_alpha[..., None] + bg[..., :-1] * np.ones(
                img_alpha.shape
            )[..., None] * (1 - img_alpha[..., None])

            return rgb.astype(np.uint8)

        return snapshot

    def export(self, uri: str | Path | bytes, **kwargs):
        """
        Use ``imageio`` to export the current Figure to a file, or return a byte string.
        Must have ``imageio`` installed.

        Parameters
        ----------
        uri: str | Path | bytes

        kwargs: passed to imageio.v3.imwrite, see: https://imageio.readthedocs.io/en/stable/_autosummary/imageio.v3.imwrite.html

        Returns
        -------
        None | bytes
            see https://imageio.readthedocs.io/en/stable/_autosummary/imageio.v3.imwrite.html
        """
        try:
            import imageio.v3 as iio
        except ModuleNotFoundError:
            raise ImportError(
                "imageio is required to use Figure.export(). Install it using pip or conda:\n"
                "pip install imageio\n"
                "conda install -c conda-forge imageio\n"
            )
        else:
            # image formats that support alpha channel:
            # https://en.wikipedia.org/wiki/Alpha_compositing#Image_formats_supporting_alpha_channels
            alpha_support = [".png", ".exr", ".tiff", ".tif", ".gif", ".jxl", ".svg"]

            uri = Path(uri)

            if uri.suffix in alpha_support:
                rgb = False
            else:
                rgb = True

            snapshot = self.export_numpy(rgb=rgb)

            return iio.imwrite(uri, snapshot, **kwargs)

    def open_popup(self, *args, **kwargs):
        warn("popups only supported by ImguiFigure")

    def _fpl_reset_layout(self, *ev):
        """set the viewport rects for all subplots, *ev argument is not used, exists because of renderer resize event"""
        self.layout.canvas_resized(self.get_pygfx_render_area())

    def get_pygfx_render_area(self, *args) -> tuple[float, float, float, float]:
        """
        Get rect for the portion of the canvas that the pygfx renderer draws to,
        i.e. non-imgui, part of canvas

        Returns
        -------
        tuple[float, float, float, float]
            x_pos, y_pos, width, height

        """

        width, height = self.canvas.get_logical_size()

        return 0, 0, width, height

    def add_subplot(
        self,
        rect=None,
        extent=None,
        camera: str | pygfx.PerspectiveCamera = "2d",
        controller: str | pygfx.Controller = None,
        name: str = None,
    ) -> Subplot:
        if isinstance(self.layout, GridLayout):
            raise NotImplementedError(
                "`add_subplot()` is not implemented for Figures using a GridLayout"
            )

        raise NotImplementedError("Not yet implemented")

        camera = create_camera(camera)
        controller = create_controller(controller, camera)

        subplot = Subplot(
            parent=self,
            camera=camera,
            controller=controller,
            canvas=self.canvas,
            renderer=self.renderer,
            name=name,
            rect=rect,
            extent=extent,  # figure created extents for grid layout
            resizeable=True,
        )

        return subplot

    def remove_subplot(self, subplot: Subplot):
        raise NotImplementedError("Not yet implemented")

        if isinstance(self.layout, GridLayout):
            raise NotImplementedError(
                "`remove_subplot()` is not implemented for Figures using a GridLayout"
            )

        if subplot not in self._subplots.tolist():
            raise KeyError(f"given subplot: {subplot} not found in the layout.")

        subplot.clear()
        self._underlay_scene.remove(subplot.frame._world_object)
        subplot.frame._world_object.clear()
        self.layout._subplots = None
        subplots = self._subplots.tolist()
        subplots.remove(subplot)
        self.layout.remove_subplot(subplot)
        del subplot

        self._subplots = np.asarray(subplots)
        self.layout._subplots = self._subplots.ravel()

    def __getitem__(self, index: str | int | tuple[int, int]) -> Subplot:
        if isinstance(index, str):
            for subplot in self._subplots.ravel():
                if subplot.name == index:
                    return subplot
            raise IndexError(f"no subplot with given name: {index}")

        if isinstance(self.layout, GridLayout):
            return self._subplots[index[0], index[1]]

        return self._subplots[index]

    def __iter__(self):
        self._current_iter = iter(range(len(self)))
        return self

    def __next__(self) -> Subplot:
        pos = self._current_iter.__next__()
        return self._subplots.ravel()[pos]

    def __len__(self):
        """number of subplots"""
        return len(self._layout)

    def __str__(self):
        return f"{self.__class__.__name__}"

    def __repr__(self):
        newline = "\n\t"

        return (
            f"fastplotlib.{self.__class__.__name__}"
            f"  Subplots:\n"
            f"\t{newline.join(subplot.__str__() for subplot in self)}"
            f"\n"
        )



# The code below is copied from https://github.com/MIERUNE/earcut-py/blob/cb30bff5458fca224c573187f36d889068ebd4e0/src/earcut/__init__.py
# which is a port of Mapbox' JS earcut (https://github.com/mapbox/earcut) version 2.2.4
# The code is not modified, except maybe formatting to keep the linter happy.
#
# ISC License
#
# Copyright (c) 2016, Mapbox
# Copyright (c) 2023, MIERUNE Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.

import math
from typing import Optional


def earcut(data, hole_indices=None, dim=2):
    has_holes = bool(hole_indices)
    outer_len = hole_indices[0] * dim if has_holes else len(data)
    outer_node = _linked_list(data, 0, outer_len, dim, True)
    triangles = []

    if (not outer_node) or outer_node.next == outer_node.prev:
        return triangles

    min_x = min_y = inv_size = None

    if has_holes:
        outer_node = _eliminate_holes(data, hole_indices, outer_node, dim)

    # if the shape is not too simple, we'll use z-order curve hash later; calculate polygon bbox
    if len(data) > 80 * dim:
        min_x = max_x = data[0]
        min_y = max_y = data[1]

        for i in range(dim, outer_len, dim):
            x = data[i]
            y = data[i + 1]
            if x < min_x:
                min_x = x
            if y < min_y:
                min_y = y
            if x > max_x:
                max_x = x
            if y > max_y:
                max_y = y

        # minX, minY and invSize are later used to transform coords into integers for z-order calculation
        inv_size = max(max_x - min_x, max_y - min_y)
        inv_size = 32767 / inv_size if inv_size != 0 else 0

    _earcut_linked(outer_node, triangles, dim, min_x, min_y, inv_size)

    return triangles


# create a circular doubly linked list from polygon points in the specified winding order
def _linked_list(data, start, end, dim, clockwise):
    last = None

    if clockwise == (_signed_area(data, start, end, dim) > 0):
        for i in range(start, end, dim):
            last = _insert_node(i, data[i], data[i + 1], last)
    else:
        for i in reversed(range(start, end, dim)):
            last = _insert_node(i, data[i], data[i + 1], last)

    if last and _equals(last, last.next):
        _remove_node(last)
        last = last.next

    return last


# eliminate colinear or duplicate points
def _filter_points(start, end=None):
    if not start:
        return start

    if not end:
        end = start

    p = start
    while True:
        again = False

        if not p.steiner and (_equals(p, p.next) or _area(p.prev, p, p.next) == 0):
            _remove_node(p)
            p = end = p.prev
            if p == p.next:
                break
            again = True

        else:
            p = p.next

        if (not again) and p == end:
            break

    return end


# main ear slicing loop which triangulates a polygon (given as a linked list)
def _earcut_linked(ear, triangles, dim, min_x, min_y, inv_size, _pass=0):
    if not ear:
        return

    # interlink polygon nodes in z-order
    if not _pass and inv_size:
        _index_curve(ear, min_x, min_y, inv_size)

    stop = ear

    # iterate through ears, slicing them one by one
    while ear.prev != ear.next:
        prev = ear.prev
        next = ear.next
        is_ear = (
            _is_ear_hashed(ear, min_x, min_y, inv_size) if inv_size else _is_ear(ear)
        )

        if is_ear:
            # cut off the triangle
            triangles.append(prev.i // dim)
            triangles.append(ear.i // dim)
            triangles.append(next.i // dim)

            _remove_node(ear)

            # skipping the next vertex leads to less sliver triangles
            ear = next.next
            stop = next.next

            continue

        ear = next

        # if we looped through the whole remaining polygon and can't find any more ears
        if ear == stop:
            # try filtering points and slicing again
            if not _pass:
                _earcut_linked(
                    _filter_points(ear), triangles, dim, min_x, min_y, inv_size, 1
                )

            # if this didn't work, try curing all small self-intersections locally
            elif _pass == 1:
                ear = _cure_local_intersections(_filter_points(ear), triangles, dim)
                _earcut_linked(ear, triangles, dim, min_x, min_y, inv_size, 2)

            # as a last resort, try splitting the remaining polygon into two
            elif _pass == 2:
                _split_earcut(ear, triangles, dim, min_x, min_y, inv_size)

            break


# check whether a polygon node forms a valid ear with adjacent nodes
def _is_ear(ear):
    a = ear.prev
    b = ear
    c = ear.next

    if _area(a, b, c) >= 0:
        return False  # reflex, can't be an ear

    # now make sure we don't have other points inside the potential ear
    ax = a.x
    ay = a.y
    bx = b.x
    by = b.y
    cx = c.x
    cy = c.y

    # triangle bbox; min & max are calculated like this for speed
    x0 = (ax if ax < cx else cx) if ax < bx else (bx if bx < cx else cx)
    y0 = (ay if ay < cy else cy) if ay < by else (by if by < cy else cy)
    x1 = (ax if ax > cx else cx) if ax > bx else (bx if bx > cx else cx)
    y1 = (ay if ay > cy else cy) if ay > by else (by if by > cy else cy)

    p = c.next
    while p != a:
        if (
            (p.x >= x0 and p.x <= x1 and p.y >= y0 and p.y <= y1)
            and _point_in_triangle(ax, ay, bx, by, cx, cy, p.x, p.y)
            and _area(p.prev, p, p.next) >= 0
        ):
            return False
        p = p.next

    return True


def _is_ear_hashed(ear, min_x, min_y, inv_size):
    a = ear.prev
    b = ear
    c = ear.next

    if _area(a, b, c) >= 0:
        return False  # reflex, can't be an ear

    ax = a.x
    ay = a.y
    bx = b.x
    by = b.y
    cx = c.x
    cy = c.y

    # triangle bbox; min & max are calculated like this for speed
    x0 = (ax if ax < cx else cx) if ax < bx else (bx if bx < cx else cx)
    y0 = (ay if ay < cy else cy) if ay < by else (by if by < cy else cy)
    x1 = (ax if ax > cx else cx) if ax > bx else (bx if bx > cx else cx)
    y1 = (ay if ay > cy else cy) if ay > by else (by if by > cy else cy)

    # z-order range for the current triangle bbox
    min_z = _z_order(x0, y0, min_x, min_y, inv_size)
    max_z = _z_order(x1, y1, min_x, min_y, inv_size)

    p = ear.prev_z
    n = ear.next_z

    # look for points inside the triangle in both directions
    while p and p.z >= min_z and n and n.z <= max_z:
        if (
            (p.x >= x0 and p.x <= x1 and p.y >= y0 and p.y <= y1)
            and (p != a and p != c)
            and _point_in_triangle(ax, ay, bx, by, cx, cy, p.x, p.y)
            and _area(p.prev, p, p.next) >= 0
        ):
            return False
        p = p.prev_z

        if (
            (n.x >= x0 and n.x <= x1 and n.y >= y0 and n.y <= y1)
            and (n != a and n != c)
            and _point_in_triangle(ax, ay, bx, by, cx, cy, n.x, n.y)
            and _area(n.prev, n, n.next) >= 0
        ):
            return False
        n = n.next_z

    # look for remaining points in decreasing z-order
    while p and p.z >= min_z:
        if (
            (p != ear.prev and p != ear.next)
            and _point_in_triangle(ax, ay, bx, by, cx, cy, p.x, p.y)
            and _area(p.prev, p, p.next) >= 0
        ):
            return False
        p = p.prev_z

    # look for remaining points in increasing z-order
    while n and n.z <= max_z:
        if (
            (n != ear.prev and n != ear.next)
            and _point_in_triangle(ax, ay, bx, by, cx, cy, n.x, n.y)
            and _area(n.prev, n, n.next) >= 0
        ):
            return False
        n = n.next_z

    return True


# go through all polygon nodes and cure small local self-intersections
def _cure_local_intersections(start, triangles, dim):
    p = start
    while True:
        a = p.prev
        b = p.next.next

        if (
            not _equals(a, b)
            and _intersects(a, p, p.next, b)
            and _locally_inside(a, b)
            and _locally_inside(b, a)
        ):
            triangles.append(a.i // dim)
            triangles.append(p.i // dim)
            triangles.append(b.i // dim)

            # remove two nodes involved
            _remove_node(p)
            _remove_node(p.next)

            p = start = b

        p = p.next
        if p == start:
            break

    return _filter_points(p)


# try splitting polygon into two and triangulate them independently
def _split_earcut(start, triangles, dim, min_x, min_y, inv_size):
    # look for a valid diagonal that divides the polygon into two
    a = start
    while True:
        b = a.next.next
        while b != a.prev:
            if a.i != b.i and _is_valid_diagonal(a, b):
                # split the polygon in two by the diagonal
                c = _split_polygon(a, b)

                # filter colinear points around the cuts
                a = _filter_points(a, a.next)
                c = _filter_points(c, c.next)

                # run earcut on each half
                _earcut_linked(a, triangles, dim, min_x, min_y, inv_size)
                _earcut_linked(c, triangles, dim, min_x, min_y, inv_size)
                return
            b = b.next
        a = a.next
        if a == start:
            break


# link every hole into the outer loop, producing a single-ring polygon without holes
def _eliminate_holes(data, hole_indices, outer_node, dim):
    queue = []
    _len = len(hole_indices)

    for i in range(_len):
        start = hole_indices[i] * dim
        end = hole_indices[i + 1] * dim if i < _len - 1 else len(data)
        lst = _linked_list(data, start, end, dim, False)
        if lst:
            if lst == lst.next:
                lst.steiner = True
            queue.append(_get_leftmost(lst))

    queue.sort(key=lambda i: i.x)

    # process holes from left to right
    for q_i in queue:
        outer_node = _eliminate_hole(q_i, outer_node)

    return outer_node


# find a bridge between vertices that connects hole with an outer ring and and link it
def _eliminate_hole(hole, outer_node):
    bridge = _find_hole_bridge(hole, outer_node)
    if not bridge:
        return outer_node

    bridge_reverse = _split_polygon(bridge, hole)

    _filter_points(bridge_reverse, bridge_reverse.next)
    return _filter_points(bridge, bridge.next)


# David Eberly's algorithm for finding a bridge between hole and outer polygon
def _find_hole_bridge(hole, outer_node):
    p = outer_node
    hx = hole.x
    hy = hole.y
    qx = -math.inf
    m = None

    # find a segment intersected by a ray from the hole's leftmost point to the left
    # segment's endpoint with lesser x will be potential connection point
    while True:
        px = p.x
        py = p.y
        if hy <= py and hy >= p.next.y and p.next.y != py:
            x = px + (hy - py) * (p.next.x - px) / (p.next.y - py)
            if x <= hx and x > qx:
                qx = x
                m = p if px < p.next.x else p.next
                if x == hx:
                    # hole touches outer segment; pick leftmost endpoint
                    return m
        p = p.next
        if p == outer_node:
            break

    if not m:
        return None

    # look for points inside the triangle of hole point, segment intersection and endpoint
    # if there are no points found, we have a valid connection
    # otherwise choose the point of the minimum angle with the ray as connection point

    stop = m
    mx = m.x
    my = m.y
    tan_min = math.inf

    p = m

    while True:
        px = p.x
        py = p.y
        if (hx >= px and px >= mx and hx != px) and _point_in_triangle(
            hx if hy < my else qx,
            hy,
            mx,
            my,
            qx if hy < my else hx,
            hy,
            px,
            py,
        ):
            tan = abs(hy - py) / (hx - px)  # tangential

            if _locally_inside(p, hole) and (
                tan < tan_min
                or (
                    tan == tan_min
                    and (px > m.x or (px == m.x and _sector_contains_sector(m, p)))
                )
            ):
                m = p
                tan_min = tan

        p = p.next
        if p == stop:
            break

    return m


# whether sector in vertex m contains sector in vertex p in the same coordinates
def _sector_contains_sector(m, p):
    return _area(m.prev, m, p.prev) < 0 and _area(p.next, m, m.next) < 0


# interlink polygon nodes in z-order
def _index_curve(start, min_x, min_y, inv_size):
    p = start
    while True:
        if p.z is None:
            p.z = _z_order(p.x, p.y, min_x, min_y, inv_size)
        p.prev_z = p.prev
        p.next_z = p.next
        p = p.next
        if p == start:
            break

    p.prev_z.next_z = None
    p.prev_z = None

    _sort_linked(p)


# Simon Tatham's linked list merge sort algorithm
# http://www.chiark.greenend.org.uk/~sgtatham/algorithms/listsort.html
def _sort_linked(_list):
    in_size = 1

    while True:
        p = _list
        _list = None
        tail = None
        num_merges = 0

        while p:
            num_merges += 1
            q = p
            p_size = 0
            for i in range(in_size):
                p_size += 1
                q = q.next_z
                if not q:
                    break
            q_size = in_size

            while p_size > 0 or (q_size > 0 and q):
                if p_size != 0 and (q_size == 0 or not q or p.z <= q.z):
                    e = p
                    p = p.next_z
                    p_size -= 1
                else:
                    e = q
                    q = q.next_z
                    q_size -= 1

                if tail:
                    tail.next_z = e
                else:
                    _list = e

                e.prev_z = tail
                tail = e

            p = q

        tail.next_z = None
        in_size *= 2

        if num_merges <= 1:
            break

    return _list


# z-order of a point given coords and inverse of the longer side of data bbox
def _z_order(x, y, min_x, min_y, inv_size):
    # coords are transformed into non-negative 15-bit integer range
    x = int((x - min_x) * inv_size)
    y = int((y - min_y) * inv_size)

    x = (x | (x << 8)) & 0x00FF00FF
    x = (x | (x << 4)) & 0x0F0F0F0F
    x = (x | (x << 2)) & 0x33333333
    x = (x | (x << 1)) & 0x55555555

    y = (y | (y << 8)) & 0x00FF00FF
    y = (y | (y << 4)) & 0x0F0F0F0F
    y = (y | (y << 2)) & 0x33333333
    y = (y | (y << 1)) & 0x55555555

    return x | (y << 1)


# find the leftmost node of a polygon ring
def _get_leftmost(start):
    p = start
    leftmost = start

    while True:
        if p.x < leftmost.x or (p.x == leftmost.x and p.y < leftmost.y):
            leftmost = p

        p = p.next
        if p == start:
            break

    return leftmost


# check if a point lies within a convex triangle
def _point_in_triangle(ax, ay, bx, by, cx, cy, px, py):
    pax = ax - px
    pay = ay - py
    pbx = bx - px
    pby = by - py
    pcx = cx - px
    pcy = cy - py
    return (
        pcx * pay - pax * pcy >= 0
        and pax * pby - pbx * pay >= 0
        and pbx * pcy - pcx * pby >= 0
    )


# check if a diagonal between two polygon nodes is valid (lies in polygon interior)
def _is_valid_diagonal(a, b):
    return (
        # dones't intersect other edges
        (a.next.i != b.i and a.prev.i != b.i and not _intersects_polygon(a, b))
        and (
            # locally visible
            (_locally_inside(a, b) and _locally_inside(b, a) and _middle_inside(a, b))
            # does not create opposite-facing sectors
            and (_area(a.prev, a, b.prev) or _area(a, b.prev, b))
            # special zero-length case
            or (
                _equals(a, b)
                and _area(a.prev, a, a.next) > 0
                and _area(b.prev, b, b.next) > 0
            )
        )
    )


# signed area of a triangle
def _area(p, q, r):
    px = p.x
    py = p.y
    qx = q.x
    qy = q.y
    rx = r.x
    ry = r.y
    return (qy - py) * (rx - qx) - (qx - px) * (ry - qy)


# check if two points are equal
def _equals(p1, p2):
    return p1.x == p2.x and p1.y == p2.y


# check if two segments intersect
def _intersects(p1, q1, p2, q2):
    o1 = _sign(_area(p1, q1, p2))
    o2 = _sign(_area(p1, q1, q2))
    o3 = _sign(_area(p2, q2, p1))
    o4 = _sign(_area(p2, q2, q1))

    if (
        (o1 != o2 and o3 != o4)  # general case
        or (
            o1 == 0 and _on_segment(p1, p2, q1)
        )  # p1, q1 and p2 are collinear and p2 lies on p1q1
        or (
            o2 == 0 and _on_segment(p1, q2, q1)
        )  # p1, q1 and q2 are collinear and q2 lies on p1q1
        or (
            o3 == 0 and _on_segment(p2, p1, q2)
        )  # p2, q2 and p1 are collinear and p1 lies on p2q2
        or (
            o4 == 0 and _on_segment(p2, q1, q2)
        )  # p2, q2 and q1 are collinear and q1 lies on p2q2
    ):
        return True

    return False


# for collinear points p, q, r, check if point q lies on segment pr
def _on_segment(p, q, r):
    return (
        q.x <= max(p.x, r.x)
        and q.x >= min(p.x, r.x)
        and q.y <= max(p.y, r.y)
        and q.y >= min(p.y, r.y)
    )


def _sign(num):
    if num > 0:
        return 1
    elif num < 0:
        return -1
    else:
        return 0


# check if a polygon diagonal intersects any polygon segments
def _intersects_polygon(a, b):
    p = a
    while True:
        pi = p.i
        ai = a.i
        bi = b.i
        pnext = p.next
        pnexti = pnext.i
        if (pi != ai and pnexti != ai and pi != bi and pnexti != bi) and _intersects(
            p, pnext, a, b
        ):
            return True

        p = pnext
        if p == a:
            break

    return False


# check if a polygon diagonal is locally inside the polygon
def _locally_inside(a, b):
    aprev = a.prev
    anext = a.next
    if _area(aprev, a, anext) < 0:
        return _area(a, b, anext) >= 0 and _area(a, aprev, b) >= 0
    else:
        return _area(a, b, aprev) < 0 or _area(a, anext, b) < 0


# check if the middle point of a polygon diagonal is inside the polygon
def _middle_inside(a, b):
    p = a
    inside = False
    px = (a.x + b.x) / 2
    py = (a.y + b.y) / 2
    while True:
        p_x = p.x
        p_y = p.y
        p_next = p.next
        p_next_y = p_next.y
        if (
            (p_y > py) != (p_next_y > py)
            and p_next.y != p_y
            and (px < (p_next.x - p_x) * (py - p_y) / (p_next_y - p_y) + p_x)
        ):
            inside = not inside
        p = p_next
        if p == a:
            break

    return inside


# link two polygon vertices with a bridge; if the vertices belong to the same ring, it splits polygon into two
# if one belongs to the outer ring and another to a hole, it merges it into a single ring
def _split_polygon(a, b):
    a2 = _Node(a.i, a.x, a.y)
    b2 = _Node(b.i, b.x, b.y)
    an = a.next
    bp = b.prev

    a.next = b
    b.prev = a

    a2.next = an
    an.prev = a2
    b2.next = a2
    a2.prev = b2
    bp.next = b2
    b2.prev = bp

    return b2


# create a node and optionally link it with previous one (in a circular doubly linked list)
def _insert_node(i, x, y, last):
    p = _Node(i, x, y)

    if not last:
        p.prev = p
        p.next = p

    else:
        p.next = last.next
        p.prev = last
        last.next.prev = p
        last.next = p

    return p


def _remove_node(p):
    p.next.prev = p.prev
    p.prev.next = p.next

    if p.prev_z:
        p.prev_z.next_z = p.next_z

    if p.next_z:
        p.next_z.prev_z = p.prev_z


class _Node:
    __slots__ = ["i", "x", "y", "prev", "next", "z", "prev_z", "next_z", "steiner"]
    i: int
    x: float
    y: float
    prev: Optional["_Node"]
    next: Optional["_Node"]
    z: Optional[int]
    prev_z: Optional["_Node"]
    next_z: Optional["_Node"]
    steiner: bool

    def __init__(self, i, x, y):
        # vertex index in coordinates array
        self.i = i

        # vertex coordinates
        self.x = x
        self.y = y

        # previous and next vertex nodes in a polygon ring
        self.prev = None
        self.next = None

        # z-order curve value
        self.z = None

        # previous and next nodes in z-order
        self.prev_z = None
        self.next_z = None

        # indicates whether this is a steiner point
        self.steiner = False


def _signed_area(data, start, end, dim):
    sum = 0
    j = end - dim
    for i in range(start, end, dim):
        sum += (data[j] - data[i]) * (data[i + 1] + data[j + 1])
        j = i

    return sum


# return a percentage difference between the polygon area and its triangulation area
# used to verify correctness of triangulation
def deviation(data, hole_indices, dim, triangles):
    has_holes = hole_indices and len(hole_indices)
    outer_len = hole_indices[0] * dim if has_holes else len(data)

    polygon_area = abs(_signed_area(data, 0, outer_len, dim))
    if has_holes:
        _len = len(hole_indices)
        for i in range(_len):
            start = hole_indices[i] * dim
            end = hole_indices[i + 1] * dim if i < _len - 1 else len(data)
            polygon_area -= abs(_signed_area(data, start, end, dim))

    triangles_area = 0
    for i in range(0, len(triangles), 3):
        a = triangles[i] * dim
        b = triangles[i + 1] * dim
        c = triangles[i + 2] * dim
        triangles_area += abs(
            (data[a] - data[c]) * (data[b + 1] - data[a + 1])
            - (data[a] - data[b]) * (data[c + 1] - data[a + 1])
        )

    if polygon_area == 0 and triangles_area == 0:
        return 0
    return abs((triangles_area - polygon_area) / polygon_area)


# turn a polygon in a multi-dimensional array form (e.g. as in GeoJSON) into a form Earcut accepts
def flatten(data):
    dim = len(data[0][0])
    vertices = []
    holes = []
    hole_index = 0

    for i in range(len(data)):
        for j in range(len(data[i])):
            for d in range(dim):
                vertices.append(data[i][j][d])

        if i > 0:
            hole_index += len(data[i - 1])
            holes.append(hole_index)

    return (vertices, holes, dim)



from typing import *
from dataclasses import dataclass
from functools import partial

import numpy as np
import pygfx

from pygfx import WorldObject, Line, Mesh, Points

from .._base import Graphic


@dataclass
class MoveInfo:
    """
    stores move info for a WorldObject
    """

    # The initial selection. Differs per type of selector
    start_selection: Any

    # The initial world position of the cursor
    start_position: np.ndarray | None

    # Delta position in world coordinates
    delta: np.ndarray

    # WorldObject or "key" event
    source: WorldObject | str


# key bindings used to move the selector
key_bind_direction = {
    "ArrowRight": np.array([1, 0, 0]),
    "ArrowLeft": np.array([-1, 0, 0]),
    "ArrowUp": np.array([0, 1, 0]),
    "ArrowDown": np.array([0, -1, 0]),
}


# Selector base class
class BaseSelector(Graphic):
    @property
    def axis(self) -> str:
        return self._axis

    @property
    def fill_color(self) -> pygfx.Color:
        """Returns the fill color of the selector, ``None`` if selector has no fill."""
        return self._fill_color

    @fill_color.setter
    def fill_color(self, color: str | Sequence[float]):
        """
        Set the fill color of the selector.

        Parameters
        ----------
        color : str | Sequence[float]
            String or sequence of floats that gets converted into a ``pygfx.Color`` object.
        """
        color = pygfx.Color(color)
        for fill in self._fill:
            fill.material.color = color
            self._original_colors[fill] = color
        self._fill_color = color

    @property
    def vertex_color(self) -> pygfx.Color:
        """Returns the vertex color of the selector, ``None`` if selector has no vertices."""
        return self._vertex_color

    @vertex_color.setter
    def vertex_color(self, color: str | Sequence[float]):
        """
        Set the vertex color of the selector.

        Parameters
        ----------
        color : str | Sequence[float]
            String or sequence of floats that gets converted into a ``pygfx.Color`` object.
        """
        color = pygfx.Color(color)
        for vertex in self._vertices:
            vertex.material.color = color
            vertex.material.edge_color = color
            self._original_colors[vertex] = color
        self._vertex_color = color

    @property
    def edge_color(self) -> pygfx.Color:
        """Returns the edge color of the selector"""
        return self._edge_color

    @edge_color.setter
    def edge_color(self, color: str | Sequence[float]):
        """
        Set the edge color of the selector.

        Parameters
        ----------
        color : str | Sequence[float]
            String or sequence of floats that gets converted into a ``pygfx.Color`` object.
        """
        color = pygfx.Color(color)
        for edge in self._edges:
            edge.material.color = color
            self._original_colors[edge] = color
        self._edge_color = color

    def __init__(
        self,
        edges: Tuple[Line, ...] = None,
        fill: Tuple[Mesh, ...] = None,
        vertices: Tuple[Points, ...] = None,
        hover_responsive: Tuple[WorldObject, ...] = None,
        arrow_keys_modifier: str = None,
        axis: str = None,
        parent: Graphic = None,
        **kwargs,
    ):
        if edges is None:
            edges = tuple()

        if fill is None:
            fill = tuple()

        if vertices is None:
            vertices = tuple()

        self._edges: Tuple[Line, ...] = edges
        self._fill: Tuple[Mesh, ...] = fill
        self._vertices: Tuple[Points, ...] = vertices

        self._world_objects: Tuple[WorldObject, ...] = (
            self._edges + self._fill + self._vertices
        )

        for wo in self._world_objects:
            wo.material.pick_write = True

        self._hover_responsive: Tuple[WorldObject, ...] = hover_responsive

        # Original color of object that we change the colors of
        self._original_colors = {}

        # Colors as they are changed by the hover events, so they can be restored after a move action
        self._hover_colors = {}

        if hover_responsive is not None:
            for wo in self._hover_responsive:
                self._original_colors[wo] = wo.material.color

        self._axis = axis

        self.arrow_keys_modifier = arrow_keys_modifier
        # if not False, moves the slider on every render cycle
        self._key_move_value = False
        self.step: float = 1.0  #: step size for moving selector using the arrow keys
        self.arrow_key_events_enabled = False

        self._move_info: MoveInfo = None

        # sets to `True` on "pointer_down", sets to `False` on "pointer_up"
        self._moving = False  #: indicates if the selector is currently being moved

        self._initial_controller_state: bool = None

        # used to disable fill area events if the edge is being actively hovered
        # otherwise annoying and requires too much accuracy to move just an edge
        self._edge_hovered: bool = False

        self._pygfx_event = None

        self._parent = parent

        Graphic.__init__(self, **kwargs)

    def get_selected_index(self):
        """Not implemented for this selector"""
        raise NotImplementedError

    def get_selected_indices(self):
        """Not implemented for this selector"""
        raise NotImplementedError

    def get_selected_data(self):
        """Not implemented for this selector"""
        raise NotImplementedError

    def _get_source(self, graphic):
        if self._parent is None and graphic is None:
            raise AttributeError(
                "No Graphic to apply selector. "
                "You must either set a ``parent`` Graphic on the selector, or pass a graphic."
            )

        # use passed graphic if provided, else use parent
        if graphic is not None:
            source = graphic
        else:
            source = self._parent

        return source

    def _fpl_add_plot_area_hook(self, plot_area):
        self._plot_area = plot_area

        # when the pointer is pressed on a fill, edge or vertex
        for wo in self._world_objects:
            pfunc_down = partial(self._move_start, wo)
            wo.add_event_handler(pfunc_down, "pointer_down")

            # double-click to enable arrow-key moveable mode
            wo.add_event_handler(self._toggle_arrow_key_moveable, "double_click")

        for fill in self._fill:
            if fill.material.color.a < 1 or fill.material.opacity < 1:
                self._pfunc_fill = partial(self._check_fill_pointer_event, fill)
                self._plot_area.renderer.add_event_handler(
                    self._pfunc_fill, "pointer_down"
                )

        # when the pointer moves
        self._plot_area.renderer.add_event_handler(self._move, "pointer_move")

        # when the pointer is released
        self._plot_area.renderer.add_event_handler(self._move_end, "pointer_up")

        # move directly to location of center mouse button click
        self._plot_area.renderer.add_event_handler(self._move_to_pointer, "click")

        # mouse hover color events
        for wo in self._hover_responsive:
            wo.add_event_handler(self._pointer_enter, "pointer_enter")
            wo.add_event_handler(self._pointer_leave, "pointer_leave")

        # arrow key bindings
        self._plot_area.renderer.add_event_handler(self._key_down, "key_down")
        self._plot_area.renderer.add_event_handler(self._key_up, "key_up")
        self._plot_area.add_animations(self._key_hold)

    def _check_fill_pointer_event(self, event_source: WorldObject, ev):
        if self._edge_hovered:
            # if edge is hovered, prefer edge events, disable fill moves
            return

        world_pos = self._plot_area.map_screen_to_world(ev)
        # outside viewport, ignore
        # this shouldn't be possible since the event handler is registered to the fill mesh world object
        # but I like sanity checks anyways
        if world_pos is None:
            return

        bbox = event_source.get_world_bounding_box()

        xmin, ymin, zmin = bbox[0]
        xmax, ymax, zmax = bbox[1]

        if not (xmin <= world_pos[0] <= xmax):
            return

        if not (ymin <= world_pos[1] <= ymax):
            return

        self._move_start(event_source, ev)

    def _move_start(self, event_source: WorldObject, ev):
        """
        Called on "pointer_down" events

        Parameters
        ----------
        event_source: WorldObject
            event source, for example selection fill area ``Mesh`` an edge ``Line`` or vertex ``Points``.
            This helps keep the event source within the MoveInfo so that during "pointer_move" events (which are
            from the renderer) we know the original source of the "move action".

        ev: Event
            pygfx ``Event``

        """
        position = self._plot_area.map_screen_to_world(ev)

        self._move_info = MoveInfo(
            start_selection=None,
            start_position=position,
            delta=np.zeros_like(position),
            source=event_source,
        )
        self._moving = True

        self._initial_controller_state = self._plot_area.controller.enabled

    def _move(self, ev):
        """
        Called on pointer move events

        Parameters
        ----------
        ev

        Returns
        -------

        """
        if self._move_info is None:
            return

        # disable controller during moves
        self._plot_area.controller.enabled = False

        # get pointer current world position, in 'mouse capute mode'
        world_pos = self._plot_area.map_screen_to_world(ev, allow_outside=True)

        # update the delta
        self._move_info.delta = world_pos - self._move_info.start_position
        self._pygfx_event = ev

        self._move_graphic(self._move_info)

        # restore the initial controller state
        # if it was disabled, keep it disabled
        self._plot_area.controller.enabled = self._initial_controller_state

    def _move_graphic(self, move_info: MoveInfo):
        raise NotImplementedError("Must be implemented in subclass")

    def _move_end(self, ev):
        self._move_info = None
        self._moving = False

        # Reset hover state
        for wo, color in self._hover_colors.items():
            wo.material.color = color
        self._hover_colors.clear()

        # restore the initial controller state
        # if it was disabled, keep it disabled
        if self._initial_controller_state is not None:
            self._plot_area.controller.enabled = self._initial_controller_state

    def _move_to_pointer(self, ev):
        """
        Calculates delta just using current world object position and calls self._move_graphic().
        """
        # check for middle mouse button click
        if ev.button != 3:
            return

        if self.axis == "x":
            offset = self.offset[0]
        elif self.axis == "y":
            offset = self.offset[1]

        if self.selection.size > 1:
            # linear region selectors
            # TODO: get center for rectangle and polygon selectors
            center = self.selection.mean(axis=0)

        else:
            # linear selectors
            center = self.selection

        current_pos_world: np.ndarray = center + offset

        world_pos = self._plot_area.map_screen_to_world(ev)

        # outside this viewport
        if world_pos is None:
            return

        delta = world_pos - current_pos_world
        self._pygfx_event = ev

        # use fill by default as the source, such as in region selectors
        if len(self._fill) > 0:
            move_info = MoveInfo(
                start_selection=None,
                start_position=None,
                delta=delta,
                source=self._fill[0],
            )
        # else use an edge, such as for linear selector
        else:
            move_info = MoveInfo(
                start_position=None,
                start_selection=None,
                delta=delta,
                source=self._edges[0],
            )

        self._move_graphic(move_info)

    def _pointer_enter(self, ev):
        if self._hover_responsive is None:
            return

        wo = ev.pick_info["world_object"]
        if wo not in self._hover_responsive:
            return

        if wo in self._edges:
            self._edge_hovered = True

        if self._moving:
            self._hover_colors[wo] = "magenta"
        else:
            wo.material.color = "magenta"

    def _pointer_leave(self, ev):
        if self._hover_responsive is None:
            return

        self._edge_hovered = False

        # reset colors
        for wo in self._hover_responsive:
            if self._moving:
                self._hover_colors[wo] = self._original_colors[wo]
            else:
                wo.material.color = self._original_colors[wo]

    def _toggle_arrow_key_moveable(self, ev):
        self.arrow_key_events_enabled = not self.arrow_key_events_enabled

    def _key_hold(self):
        if self._key_move_value and self.arrow_key_events_enabled:
            # direction vector * step
            delta = key_bind_direction[self._key_move_value] * self.step

            # set event source
            # use fill by default as the source
            if len(self._fill) > 0:
                move_info = MoveInfo(
                    start_selection=None,
                    start_position=None,
                    delta=delta,
                    source=self._fill[0],
                )
            # else use an edge
            else:
                move_info = MoveInfo(
                    start_selection=None,
                    start_position=None,
                    delta=delta,
                    source=self._edges[0],
                )

            # move the graphic
            self._move_graphic(move_info)

    def _key_down(self, ev):
        # key bind modifier must be set and must be used for the event
        # for example. if "Shift" is set as a modifier, then "Shift" must be used as a modifier during this event
        if (
            self.arrow_keys_modifier is not None
            and self.arrow_keys_modifier not in ev.modifiers
        ):
            return

        # ignore if non-arrow key is pressed
        if ev.key not in key_bind_direction.keys():
            return

        self._key_move_value = ev.key

    def _key_up(self, ev):
        # if arrow key is released, stop moving
        if ev.key in key_bind_direction.keys():
            self._key_move_value = False

    def _fpl_prepare_del(self):
        if hasattr(self, "_pfunc_fill"):
            self._plot_area.renderer.remove_event_handler(
                self._pfunc_fill, "pointer_down"
            )
            del self._pfunc_fill
        super()._fpl_prepare_del()
